from collections import OrderedDict, defaultdict
from contextlib import contextmanager
from typing import Dict, List

from torch import Tensor

from . import metric


def t2float(value):
    if isinstance(value, Tensor) and value.dim() == 0:
        return value.item()
    return value


class MeterInterface:
    """
    meter interface only concerns about the situation in one epoch,
    without considering historical record and save/load state_dict function.
    """

    def __init__(self, default_focus="tra") -> None:
        self._group_bank: Dict[str, Dict[str, metric.Metric]] = defaultdict(OrderedDict)
        self.__group_focus = default_focus

    def register_meter(self, name: str, meter: metric.Metric):
        return self._register_meter(name=name, meter=meter, group_name=self.__group_focus)

    def _register_meter(self, *, name: str, group_name: str, meter: metric.Metric, ) -> None:
        if not isinstance(meter, metric.Metric):
            raise KeyError(meter)
        group_meter = self._group_bank[group_name]
        if name in group_meter:
            raise KeyError(f"{name} exists in {group_name}")
        group_meter[name] = meter

    def _delete_meter(self, *, name: str, group_name: str) -> None:
        meters = self._get_meters_by_group(group_name=group_name)
        if name not in meters:
            raise KeyError(name)
        del self._group_bank[group_name][name]
        if len(self._group_bank[group_name]) == 0:
            del self._group_bank[group_name]

    def delete_meter(self, name: str):
        return self._delete_meter(name=name, group_name=self.__group_focus)

    def delete_meters(self, name_list: List[str]):
        for name in name_list:
            self.delete_meter(name=name)

    def add(self, meter_name, *args, **kwargs):
        meter = self._get_meter(name=meter_name, group_name=self.__group_focus)
        meter.add(*args, **kwargs)

    def reset(self) -> None:
        for g in self.groups():
            for m in self._group_bank[g].values():
                m.reset()

    def join(self):
        for g in self.groups():
            meters = self._get_meters_by_group(g)
            for m in meters.values():
                m.join()

    def _get_meters_by_group(self, group_name: str):
        if group_name not in self.groups():
            raise KeyError(f"{group_name} not in {self.__class__.__name__}: ({', '.join(self.groups())})")
        return self._group_bank[group_name]

    def _get_meter(self, *, name: str, group_name: str):
        meters: Dict[str, metric.Metric] = self._get_meters_by_group(group_name=group_name)
        if name not in meters:
            raise KeyError(f"{name} not in {group_name} group: ({', '.join(meters)})")
        return meters[name]

    def groups(self):
        return list(self._group_bank.keys())

    @property
    def cur_focus(self):
        return self.__group_focus

    @contextmanager
    def focus_on(self, group_name: str):
        prev_focus = self.__group_focus
        self.__group_focus = group_name
        yield
        self.__group_focus = prev_focus

    def _statistics_by_group(self, group_name: str):
        meters = self._get_meters_by_group(group_name)
        return {k: t2float(m.summary()) for k, m in meters.items()}

    def statistics(self):
        """get statistics from meter_interface. ignoring the group with name starting with `_`"""
        groups = self.groups()
        for g in groups:
            if not g.startswith("_"):
                yield g, self._statistics_by_group(g)

    def __enter__(self):
        self.reset()

    def __exit__(self, *args, **kwargs):
        self.join()

    def __getitem__(self, meter_name: str) -> metric.Metric:
        return self._get_meter(name=meter_name, group_name=self.__group_focus)



import torch
import typing as t
from collections import defaultdict
from torch import Tensor
from .metric import Metric, DistributedMixin
from torch import distributed as dist

metric_result = Tensor
dictionary_metric_result = Metric[t.Union[str, metric_result]]


class AverageValueMeter(DistributedMixin, Metric[metric_result]):
    def __init__(self):
        super(AverageValueMeter, self).__init__()
        self.sum = torch.tensor(0.0)
        self.n = 0

    @t.overload
    def add(self, value: Tensor, n: int = 1):
        ...

    def add(self, *args, **kwargs):
        return super(AverageValueMeter, self).add(*args, **kwargs)

    def _add(self, value: Tensor, n=1):
        device, dtype = value.device, value.dtype
        self.sum = self.sum.to(device=device, dtype=dtype) + value.detach() * n
        self.n += n

    def _synchronize(self):
        dist.all_reduce(self.sum, op=dist.ReduceOp.SUM)
        self.sum /= self.process_num

    def reset(self):
        self.sum = torch.tensor(0.0)
        self.n = 0

    def _summary(self) -> metric_result:
        # this function returns a dict and tends to aggregate the historical results.
        return torch.nan if self.n == 0 else self.sum / self.n  # noqa


class AverageValueDictionaryMeter(DistributedMixin, Metric[dictionary_metric_result]):
    def _synchronize(self):
        for meters in self._meter_dicts.values():
            meters.synchronize()

    def __init__(self) -> None:
        super().__init__()
        self._meter_dicts: t.Dict[str, AverageValueMeter] = defaultdict(AverageValueMeter)

    def reset(self):
        for k, v in self._meter_dicts.items():
            v.reset()

    def _add(self, **kwargs):
        for k, v in kwargs.items():
            self._meter_dicts[k].add(v)

    def _summary(self):
        return {k: v.summary() for k, v in self._meter_dicts.items()}


class AverageValueListMeter(AverageValueDictionaryMeter):

    def _add(self, list_value: t.Iterable[float] = None, **kwargs):
        assert isinstance(list_value, t.Iterable)
        for i, v in enumerate(list_value):
            self._meter_dicts[str(i)].add(v)



import typing as t
from abc import abstractmethod
from torch import distributed as dist

RETURN_TYPE = t.TypeVar("RETURN_TYPE")


class Metric(t.Generic[RETURN_TYPE]):
    _initialized = False

    def __init__(self, **kwargs) -> None:
        self._initialized = True

    @abstractmethod
    def reset(self):
        pass

    def add(self, *args, **kwargs):
        assert self._initialized, f"{self.__class__.__name__} must be initialized by overriding __init__"
        return self._add(*args, **kwargs)

    @abstractmethod
    def _add(self, *args, **kwargs):
        pass

    def summary(self) -> RETURN_TYPE:
        return self._summary()

    @abstractmethod
    def _summary(self) -> RETURN_TYPE:
        pass

    @t.final
    def join(self):
        return

    @t.final
    def close(self):
        return


BASE = Metric if t.TYPE_CHECKING else object


class DistributedMixin(BASE):
    @abstractmethod
    def _synchronize(self):
        """Synchronize values across GPUs"""
        pass

    @t.final
    def synchronize(self):
        if self.is_distributed:
            self._synchronize()

    @property
    def is_distributed(self) -> bool:
        return dist.is_initialized()

    @property
    def process_num(self) -> int:
        try:
            return dist.get_world_size()
        except:  # noqa
            return 1

    def add(self, *args, **kwargs):
        super(DistributedMixin, self).add(*args, **kwargs)
        self.synchronize()
